# import dill
import math
from .constants import *
from .layers import *
from .utils import *
import bson
import gdsfactory as gf
from copy import deepcopy
# from time import time
import datetime
import json
import subprocess
import sys
import time
from functools import partial
from math import cos, pi, sin
import os
import numpy as np

from gdsfactory.cross_section import Section
from sortedcontainers import SortedDict, SortedSet
from sympy import N
from gdsfactory.generic_tech import LAYER_STACK, LAYER


def setup(c, study,   dx, margin,
          bbox_layer=LAYER.WAFER,
          zmargin=None, zlims=None, core_layer=LAYER.WG,
          port_source_offset="auto", source_margin="auto", name="",
          runs=[],  sources=[],
          layer_stack=LAYER_STACK, materials=MATERIALS,
          exclude_layers=[
              DESIGN_LAYER, GUESS], approx_2D=False, Courant=None,
          gpu=None, dtype=np.float32,
          path=RUNS_PATH, plot=False, **kwargs):
    if type(bbox_layer[0]) is int:
        bbox_layer = (bbox_layer,)
    prob = dict()
    prob = {**prob, **kwargs}
    prob["dtype"] = str(dtype)
    prob["timestamp"] = datetime.datetime.now().isoformat(
        timespec="seconds").replace(":", "-")
    prob["name"] = name
    prob["gpu_backend"] = gpu if gpu else ""
    ports = {
        p.name: {
            "center": p.center,
            "normal": [cos(p.orientation/180*pi), sin(p.orientation/180*pi)],
        }
        for p in c.get_ports_list(prefix="o")
    }
    prob["ports"] = ports

    mode_solutions = []

    if zlims is None:
        # if core not in layer_stack.layers:
        #     raise ValueError("please set core=\"core_layer_name\"")
        # "\"core\" not found in layer_stack so `zlims` must be provided")
        d = get_layers(layer_stack, core_layer)[0]
        hcore = d.thickness
        zmin = d.zmin
        # hcore = round(d.thickness/dx)*dx
        zlims = [zmin, zmin+hcore]
    thickness = zlims[1]-zlims[0]

    a = min([min([materials[d.material].epsilon for d in get_layers(
        layer_stack, l)]) for l in bbox_layer])
    b = max([materials[d.material].epsilon for d in get_layers(
        layer_stack, core_layer)])
    C = 5*math.sqrt(a/b)
    if zmargin is None:
        zmargin = dx*round(C*thickness/dx)
    port_width = max([p.width/1e3 for p in c.ports])
    if margin is None:
        margin = trim(C*port_width, dx)

    c = add_bbox(c, layer=bbox_layer, nonport_margin=margin)
    layers = set(c.layers)-set(exclude_layers)

    # zlims = [zlims[0]-dx, zlims[0]+np.ceil(thickness/dx)*dx+dx]
    zlims = [zlims[0]-zmargin, zlims[1]+zmargin]
    h = zlims[1]-zlims[0]
    zcenter = zlims[0]+(zlims[1]-zlims[0])/2
    zmin = zlims[0]

    l, w = c.bbox_np()[1]-c.bbox_np()[0]
    # l, w = round(l/dx)*dx, round(w/dx)*dx
    center = [c.bbox_np()[0][0], c.bbox_np()[0][1]+w/2, zcenter]
    normal = [1, 0, 0]

    eps = material_voxelate(c, dx, center, l, w, h,
                            normal, layers, layer_stack, materials)
    eps_2D = eps[:, :, int(eps.shape[2]/2)]
    prob["study"] = study

    l = [prob["timestamp"], study]
    if name:
        l.append(name)
    path = os.path.join(path, "#".join(l))
    prob["path"] = path

    prob["eps_3D"] = eps.tolist()
    prob["eps"] = prob["eps_3D"]
    prob["eps_2D"] = eps_2D.tolist()
    prob["zmin"] = zmin
    prob["d"] = 2 if approx_2D else 3

    prob["mode_height"] = h
    w = port_width+2*margin
    neffmin = 1000000
    wavelengths = []
    # _c = add_bbox(c, layer=bbox_layer, nonport_margin=margin)
    for run in runs:
        for s in list(run["sources"].values())+list(run["monitors"].values()):
            s["mode_width"] = w
            for wl in s["wavelength_mode_numbers"]:
                wavelengths.append(wl)
                duplicate = False
                mode_numbers = s["wavelength_mode_numbers"][wl]
                for m in mode_solutions:
                    if m["wavelength"] == wl and max(mode_numbers) < len(m["modes"]) and m["width"] == s["width"]:
                        m["ports"].append(s["port"])
                        duplicate = True
                if not duplicate:
                    normal = s["normal"]+[0]
                    center = list(s["center"])+[zcenter]

                    center = np.array(center)-.001*np.array(normal)
                    eps = material_slice(
                        c, dx, center, w, h, normal, layers, layer_stack, materials)
                    _modes, _modes1, neffs, _ = solve_modes(
                        eps, λ=wl, dx=dx, neigs=max(mode_numbers)+1, plot=plot)

                    for n in neffs:
                        if n < neffmin:
                            neffmin = n

                    modes = [{k: [np.real(mode[k].T).tolist(), np.imag(
                        mode[k].T).tolist()] for k in mode} for mode in _modes]
                    modes1 = [{k: [np.real(mode[k]).tolist(), np.imag(
                        mode[k]).tolist()] for k in mode} for mode in _modes1]
                    # for mn in mode_numbers:
                    mode_solutions.append({
                        "modes": modes,
                        "modes1": modes1,
                        "wavelength": wl,
                        # "mode_number": mn,
                        "ports": [s["port"]],
                        "size": [w, h],
                        "eps": eps.tolist(),
                        "width": s["width"],
                        "mode_width": w,
                        "zcenter": zcenter,
                    })
    wavelengths = sorted(set(wavelengths))
    wl = np.median(wavelengths)
    if port_source_offset == "auto":
        port_source_offset = trim(3*wl/neffmin, dx)
    prob["port_source_offset"] = port_source_offset
    if source_margin == "auto":
        source_margin = 2*dx
    prob["source_margin"] = source_margin

    prob["margin"] = margin
    prob["zmargin"] = zmargin
    prob["portsides"] = portsides(c)

    prob["mode_solutions"] = mode_solutions
    prob["runs"] = runs
    # device_svg = write_img("device", c,
    #                        hidden_layer=(DESIGN, GUESS))
    prob["components"] = {
        "device": {
            # "svg": device_svg,
            "bbox": c.bbox_np().tolist(),
        }}
    prob["dx"] = dx
    prob["Courant"] = Courant
    # prob["component"] = c
    if not os.path.exists(path):
        os.makedirs(path)
    # with open(os.path.join(path, "comp.bson"), "ab") as f:
    #     dill.dump(c, f)
    # λc = (wavelengths[0]+wavelengths[-1])/2
    # prob["wavelengths"] = wavelengths
    prob["mode_solutions"] = mode_solutions
    # prob["path_length"] = 2*(l+r+lwg)
    return prob


def port_number(port):
    s = str(port).split("@")[0]
    if s[0] == "o":
        s = s[1:]
    return int(s)


def mode_number(port):
    l = str(port).split("@")
    return 0 if len(l) == 1 else int(l[1])


def longname(s):
    o, i = s.split(",")
    po, pi = port_number(o), port_number(i)
    mo, mi = mode_number(o), mode_number(i)
    return f"o{po}@{mo},o{pi}@{mi}"



# -*- coding: utf-8 -*-
"""Convert an image file to a GDS file
"""

import cv2
import numpy as np
import gdspy

import argparse


def minmax(v):
    if v > 255:
        v = 255
    if v < 0:
        v = 0
    return v


def main(fileName, sizeOfTheCell, layerNum, isDither, scale):
    """Convert an image file (fileName) to a GDS file
    """
    print("Converting an image file to a GDS file..")
    # Read an image file
    img = cv2.resize(cv2.imread(fileName), dsize=None, fx=scale, fy=scale)

    width = img.shape[1]
    height = img.shape[0]
    print("width:{0}".format(width))
    print("height:{0}".format(height))

    # Convert an image to grayscale one
    gray = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)

    if (isDither):
        # Floyd–Steinberg dithering
        # https://en.wikipedia.org/wiki/Floyd%E2%80%93Steinberg_dithering
        for y in range(0, height-1):
            for x in range(1, width-1):
                old_p = gray[y, x]
                new_p = np.round(old_p/255.0) * 255
                gray[y, x] = new_p
                error_p = old_p - new_p
                gray[y, x+1] = minmax(gray[y, x+1] + error_p * 7 / 16.0)
                gray[y+1, x-1] = minmax(gray[y+1, x-1] + error_p * 3 / 16.0)
                gray[y+1, x] = minmax(gray[y+1, x] + error_p * 5 / 16.0)
                gray[y+1, x+1] = minmax(gray[y+1, x+1] + error_p * 1 / 16.0)

        ret, binaryImage = cv2.threshold(gray, 0, 255, cv2.THRESH_OTSU)
    else:
        ret, binaryImage = cv2.threshold(gray, 0, 255, cv2.THRESH_OTSU)

    # Fill orthological corner
    for x in range(width - 1):
        for y in range(height - 1):
            if binaryImage.item(y, x) == 0 and binaryImage.item(y + 1, x) == 255 \
                    and binaryImage.item(y, x + 1) == 255 and binaryImage.item(y + 1, x + 1) == 0:
                binaryImage.itemset((y + 1, x), 0)
            elif binaryImage.item(y, x) == 255 and binaryImage.item(y + 1, x) == 0 \
                    and binaryImage.item(y, x + 1) == 0 and binaryImage.item(y + 1, x + 1) == 255:
                binaryImage.itemset((y + 1, x + 1), 0)

    # Output image.bmp
    cv2.imwrite("image.bmp", binaryImage)

    # The GDSII file is called a library, which contains multiple cells.
    lib = gdspy.GdsLibrary()
    gdspy.current_library = gdspy.GdsLibrary()

    # Geometry must be placed in cells.
    unitCell = lib.new_cell('CELL')
    square = gdspy.Rectangle((0.0, 0.0), (1.0, 1.0), layer=(int)(layerNum))
    unitCell.add(square)

    grid = lib.new_cell("GRID")

    for x in range(width):
        for y in range(height):
            if binaryImage.item(y, x) == 0:
                # print("({0}, {1}) is black".format(x, y))
                cell = gdspy.CellReference(
                    unitCell, origin=(x, height - y - 1))
                grid.add(cell)

    scaledGrid = gdspy.CellReference(
        grid, origin=(0, 0), magnification=(float)(sizeOfTheCell))

    # Add the top-cell to a layout and save
    top = lib.new_cell("TOP")
    top.add(scaledGrid)
    s = fileName.split(".")
    s = (".").join(s[:-1])
    lib.write_gds(f"{s}.gds")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('fileName', type=str,
                        help='name of the input image file')
    parser.add_argument('sizeOfTheCell', type=float,
                        help='size of the unit-cells (minimum width and space) [um]')
    parser.add_argument('layerNum', type=int,
                        help='layer number of the output GDSII file')
    parser.add_argument('--scale', default=1.0, type=float, help='scale')
    parser.add_argument('-d', action='store_true',
                        help='Floyd–Steinberg dithering')
    args = parser.parse_args()

    main(args.fileName, args.sizeOfTheCell, args.layerNum, args.d, args.scale)



from .materials import MATERIALS
from .sparams import *
from .setup import *
from .constants import *
from .layers import *
from .utils import *
import scipy as sp
import bson
import gdsfactory as gf
from copy import deepcopy
# from time import time
import datetime
import json
import subprocess
from functools import partial
from math import cos, pi, sin
import os
from gdsfactory.generic_tech import LAYER_STACK, LAYER


def inverse_design_problem(c,  lmin=.1, symmetries=[],
                           tparam_targets={}, sparam_targets={},
                           iters=25, eta=.1, init=None,  # minloss=.01,
                           design_region_layer=DESIGN_LAYER,
                           #    design_guess_layer=LAYER.GUESS,
                           fill_layer=LAYER.WG,
                           void_layer=None,
                           layer_stack=LAYER_STACK, materials=MATERIALS,
                           contrast=10,
                           plot=False, approx_2D=True, **kwargs):
    design_region_layer = tuple(design_region_layer)
    if not approx_2D:
        raise NotImplementedError(
            "3D inverse design feature must be requested from Luminescent AI info@luminescentai.com")

    if tparam_targets:
        target_type = "tparams"
        targets = tparam_targets
    elif sparam_targets:
        target_type = "sparams"
        targets = sparam_targets

    targets = {
        wl: {
            longname(k): v for k, v in d.items()
        } for wl, d in targets.items()}
    keys = sorted(set(sum([list(d.keys()) for d in targets.values()], [])))
    wavelengths = sorted(targets.keys())

    prob = sparams_problem(c,
                           layer_stack=layer_stack, materials=materials,
                           wavelengths=wavelengths,
                           study="inverse_design",
                           keys=keys,
                           approx_2D=approx_2D, ** kwargs)
    prob["targets"] = targets
    prob["wavelengths"] = wavelengths
    prob["target_type"] = target_type
    prob["contrast"] = contrast
    # prob["init"] = init
    prob = {**prob, **kwargs}
    prob["eta"] = eta
    polys = c.extract([design_region_layer]).get_polygons()
    if not symmetries:
        symmetries = [[]]*len(polys)
    else:
        if type(symmetries[0]) is not list:
            symmetries = [symmetries]*len(polys)

    def _bbox(b):
        return [[b.left/1e3, b.bottom/1e3], [b.right/1e3, b.top/1e3]]
    prob["designs"] = [
        {
            "layer": design_region_layer,
            "bbox": _bbox(p.bbox()),
            "symmetries": s,
            "init": init,
            "lmin": lmin,
        } for p, s in zip(list(polys.values())[0], symmetries)
    ]
    epsmin = np.min(prob["eps_2D"])

    prob["design_config"] = dict()
    l = get_layers(layer_stack, fill_layer)[0]
    d = {"thickness": l.thickness, "material": l.material, "zmin": l.zmin}
    d["epsilon"] = materials[d["material"]].epsilon
    d["ϵ"] = d["epsilon"]
    prob["design_config"]["fill"] = d

    if void_layer is not None:
        l = get_layers(layer_stack, void_layer)[0]
        d = {"thickness": l.thickness, "material": l.material, "zmin": l.zmin}
        d["epsilon"] = materials[d["material"]].epsilon
        d["ϵ"] = d["epsilon"]
    else:
        d = copy.deepcopy(d)
        d["epsilon"] = epsmin
        d["ϵ"] = d["epsilon"]
    prob["design_config"]["void"] = d

    prob["design_config"]["design_region_layer"] = design_region_layer
    prob["iters"] = iters
    return prob


def apply_design(c0,  sol):
    path = sol["path"]
    a = gf.Component()
    a.add_ref(c0)
    fill = sol["design_config"]["fill_layer"]
    dl = sol["design_config"]["design_region_layer"]
    for i, d in enumerate(sol["designs"]):
        x0, y0 = d["bbox"][0]
        x1, y1 = d["bbox"][1]
        b = gf.Component()
        b.add_polygon(
            [(x0, y0), (x1, y0), (x1, y1), (x0, y1)], layer=dl)
        a = gf.boolean(a, b, "not", layer=dl)
    c = gf.Component()
    c << a
    for layer in c0.layers:
        if layer != dl:
            c.add_ref(c0.extract([layer]))
    # c.show()
    pic2gds(os.path.join(path, f"design{i+1}.png"), sol["dx"])
    g = gf.import_gds(os.path.join(path, f"design{i+1}.gds"), "TOP",
                      read_metadata=True)
    polygons = g.get_polygons(merge=True)
    p = polygons[1][0]
    g = gf.Component()
    g.add_polygon(p, layer=fill)
    g.drotate(90)
    g = c << g
    g.xmin = x0
    g.ymin = y0
    return c



# from gcells import ubend
from operator import inv
from cv2 import solve
from regex import E
from gcells import *
from prob import *
import pcells
import gdsfactory as gf
# import picToGDS as pg
import luminescent.gplugins.luminescent.utils as utils
from generic_tech import LAYER_STACK, LAYER, LAYER_VIEWS

dx = .05
wwg = 0.4
r = .4
h = 0.2
l = 3
lmin = 0.1


c, sources, monitors, design = ubend_template(
    r,  l, wwg, dx=dx, wavelength=1.55, LAYER=LAYER)
c = add_bbox(c, layers=[LAYER.BOX, LAYER.WGCLAD], margin=.2)
prob = inverse_design_problem(
    c, lmin=lmin, dx=dx, sources=sources, monitors=monitors, design=design, layer_stack=LAYER_STACK)
solve(prob)
raise Exception("Not implemented")
utils.pic2gds("opt/ubend.png", dx)
init = gf.import_gds("image.gds", "TOP",  read_metadata=True).rotate(90)
ubend = ubend_template(r, wwg, h, l, dx=dx, dir="opt", lmin=lmin, init=init)
ubend.show()
ps = pcells.phase_shifter(l=20, wwg=wwg, nbends=4, ubend=ubend)
res = pcells.ubend_resonator(l=50, wwg=wwg, ubend=ubend)
ps.show()
res.show()



import copy
from .constants import *
import gdsfactory as gf
from gdsfactory.cross_section import Section
from .constants import RUNS_PATH
from .materials import MATERIALS
try:
    from IPython.display import display
except ImportError:
    pass

from re import T
from PIL import Image
import imageio.v3 as iio
import textwrap
from cairosvg import svg2png
from regex import F
from .picToGDS import main
import rasterio.features
import shapely
from shapely import GeometryCollection, Polygon
from shapely.ops import clip_by_rect
from shapely.plotting import plot_polygon, plot_points
from .picToGDS import main
import io
import os
import cv2
import gdstk
import pylab
import EMpy
# import EMpy as em
import numpy
from math import cos, pi, sin, tan
import trimesh
import matplotlib.pyplot as plt
import numpy as np
from gdsfactory.generic_tech import LAYER_STACK, LAYER
# from gdsfactory import LAYER_VIEWS

tol = .001


def trim(x, dx):
    return round(x/dx)*dx


def extend(endpoints, wm):
    v = endpoints[1]-endpoints[0]
    v = v/np.linalg.norm(v)
    return [
        (endpoints[0]-wm*v).tolist(), (endpoints[1]+wm*v).tolist()]


def portsides(c):
    res = [[False, False], [False, False]]
    bbox = c.bbox_np()
    xmin0, ymin0 = bbox[0]
    xmax0, ymax0 = bbox[1]
    for p in c.ports:
        x, y = np.array(p.center)/1e3
        if abs(x - xmin0) < tol:
            res[0][0] = True
        if abs(x - xmax0) < tol:
            res[1][0] = True
        if abs(y - ymin0) < tol:
            res[0][1] = True
        if abs(y - ymax0) < tol:
            res[1][1] = True
    print(res)
    return res


def add_bbox(c, layer, nonport_margin=0):  # , dx=None):
    bbox = c.bbox_np()
    xmin0, ymin0 = bbox[0]
    xmax0, ymax0 = bbox[1]
    l = xmax0-xmin0
    w = ymax0-ymin0

    # l = dx*np.ceil((xmax0-xmin0)/dx)
    # w = dx*np.ceil((ymax0-ymin0)/dx)

    # if dx is not None:
    #     if nonport_margin is None:
    #         nonport_margin = dx
    # if nonport_margin is None:
    #     nonport_margin = 0
    margin = nonport_margin
    xmin, ymin, xmax, ymax = xmin0-margin, ymin0 - \
        margin, xmin0+l+margin, ymin0+w+margin

    for p in c.ports:
        # p = c.ports[k]
        x, y = np.array(p.center)/1e3
        if abs(x - xmin0) < tol:
            xmin = x
        if abs(x - xmax0) < tol:
            xmax = x
        if abs(y - ymin0) < tol:
            ymin = y
        if abs(y - ymax0) < tol:
            ymax = y
    p = [(xmin, ymin), (xmax, ymin), (xmax, ymax), (xmin, ymax)]
    _c = gf.Component()
    _c << c

    if type(layer[0]) is int:
        layer = [layer]
    for layer in layer:
        layer = tuple(layer)
        _c.add_polygon(p, layer=layer)
    for port in c.ports:
        _c.add_port(name=port.name, port=port)
    return _c


def solve_modes(eps, λ, dx, neigs=1, plot=False):
    tol = 1e-4
    m, n = eps.shape
    # print(m, n)
    x = numpy.linspace(0.5*dx, (m-.5)*dx, m)
    y = numpy.linspace(0.5*dx, (n-.5)*dx, n)

    i = round(n/2)
    y1 = y[i:i+4]
    e = eps[:, i]
    eps1 = np.stack([e, e, e, e]).T

    def ϵfunc(x_, y_):
        m, n = len(x_), len(y_)
        return cv2.resize(eps, dsize=(n, m), interpolation=cv2.INTER_CUBIC)

    def ϵfunc1(x_, y_):
        m, n = len(x_), len(y_)
        return cv2.resize(eps1, dsize=(n, m), interpolation=cv2.INTER_CUBIC)
        # return ϵfunc(x_, y_)[i:i+len(y_), :]

    tol = 1e-6
    # plot = True
    # neigs = 2
    solver = EMpy.modesolvers.FD.VFDModeSolver(
        λ, x, y, ϵfunc,  "0000").solve(neigs, tol)
    solvera = EMpy.modesolvers.FD.VFDModeSolver(
        λ, x, y1, ϵfunc1,  "AA00").solve(2*neigs, tol)
    # solvers = EMpy.modesolvers.FD.VFDModeSolver(
    #     λ, x, y1, ϵfunc1, "SS00").solve(neigs, tol)

    # fig = pylab.figure()
    # fig.add_subplot(2, 3, 1)
    # Hy = numpy.transpose(solver.modes[-1].get_field("Hy", x, y))
    # pylab.imshow(abs(Hy))
    # fig.add_subplot(2, 3, 2)
    # Hy = numpy.transpose(solvera.modes[-1].get_field("Hy", x, y1))
    # pylab.imshow(abs(Hy))
    # fig.add_subplot(2, 3, 3)
    # Hy = numpy.transpose(solvers.modes[-1].get_field("Hy", x, y1))
    # pylab.imshow(abs(Hy))
    # fig.add_subplot(2, 3, 4)
    # e = numpy.transpose(ϵfunc(x, y))
    # pylab.imshow(e)
    # fig.add_subplot(2, 3, 5)
    # e = numpy.transpose(ϵfunc1(x, y1))
    # pylab.imshow(e)
    # pylab.show()

    modes = [{k: m.get_field(k, x, y) for k in [
        "Ex", "Ey", "Ez", "Hx", "Hy", "Hz"]} for m in solver.modes]
    neffs = [np.real(m.neff) for m in solver.modes]

    v = sorted(solvera.modes, key=lambda x: -np.abs(x.neff))

    def f(m):
        a = m.get_field("Hy", x, y1)
        a = abs(a)
        return (max(np.std(a, 1)) / np.max(a)) < .1
    v = filter(f, v)
    modes1 = [{k: m.get_field(k, x, y1)[:, 0] for k in [
        "Ex", "Ey", "Ez", "Hx", "Hy", "Hz"]} for m in v]
    neffs1 = [np.real(m.neff) for m in v]
    # "Ex", "Ey", "Ez", "Hx", "Hy", "Hz"]} for m in sorted(solvera.modes+solvers.modes, key=lambda x: -np.abs(x.neff))]
    # print(solver.modes[0].get_field("Ex", x, y).shape)
    return modes, modes1, neffs, neffs1


def s2svg(area, bbox, dx):
    # specify margin in coordinate units

    width = bbox[2] - bbox[0]
    height = bbox[3] - bbox[1]
    props = {
        'version': '1.1',
        'baseProfile': 'full',
        'width': '{width:.0f}px'.format(width=round(width/dx)),
        'height': '{height:.0f}px'.format(height=round(height/dx)),
        'viewBox': '%.1f,%.1f,%.1f,%.1f' % (bbox[0], bbox[1], width, height),
        'xmlns': 'http://www.w3.org/2000/svg',
        'xmlns:ev': 'http://www.w3.org/2001/xml-events',
        'xmlns:xlink': 'http://www.w3.org/1999/xlink',
        # "fill": "#FFFFFF",
    }

    return textwrap.dedent(r'''
        <?xml version="1.0" encoding="utf-8" ?>
        <svg {attrs:s}>
        {data:s}
        </svg>
    ''').format(
        attrs=' '.join(['{key:s}="{val:s}"'.format(
            key=key, val=props[key]) for key in props]),
        data=area.svg()
    ).strip().replace('stroke-width="2.0"', 'stroke-width="0.0"')


def write_img(f, c, hidden_layer=[]):
    pgds = os.path.join(f"{f}.gds")
    c.write_gds(pgds, with_metadata=True)
    if hidden_layer:
        if hidden_layer[0] is int:
            hidden_layer = [hidden_layer]
    psvg = os.path.join(f"{f}.svg")
    gdstk.read_gds(pgds).top_level()[0].write_svg(
        psvg,
        scaling=1,
        background="#000000",
        shape_style={k: {"fill": "none",
                                 "stroke": "none",
                                 "stroke-dasharray": "8,8"} for k in hidden_layer},
        pad=0)
    return io.open(psvg, "r").read()
    # renderPM.drawToFile(svg2rlg("tmp"), f, fmt='PNG')


def pic2gds(fileName, sizeOfTheCell, layerNum=1, isDither=False, scale=1):
    main(fileName, sizeOfTheCell, layerNum, isDither, scale)
    return "image.bmp", "image.gds"


def finish(c, name):
    c.add_label(name, position=c.bbox_np()[1])


def normal_from_orientation(orientation):
    return [cos(orientation/180*pi), sin(orientation/180*pi)]


def raster_slice(scene, dx, center, w, h, normal):
    # scene.show()
    g = sum(scene.geometry.values())
    s = g.section(plane_origin=center, plane_normal=normal)
    tangent = np.array([[-normal[1], normal[0], 0]]).T
    normal = np.array(normal)
    normal.shape = (3, 1)
    center = np.array(center)
    center.shape = (3, 1)
    tangent2 = np.cross(normal.T, tangent.T).T

    R = np.linalg.inv(np.concatenate((tangent, tangent2, normal), axis=1))
    to_2D = np.eye(4)
    to_2D[:3, :3] = R

    v = -tangent*w/2-h/2*tangent2
    xmin,    ymin, _ = numpy.reshape(np.matmul(R, v), (3,))

    v = +tangent*w/2+h/2*tangent2
    xmax,    ymax, _ = numpy.reshape(np.matmul(R, v), (3,))

    xmin, xmax = min(xmin, xmax), max(xmax, xmin)
    ymin, ymax = min(ymin, ymax), max(ymax, ymin)

    x, y, z = numpy.reshape(np.matmul(R, center), (3,))
    xmin, xmax = xmin+x, xmax+x
    ymin, ymax = ymin+y, ymax+y

    if not s:
        return np.zeros((round(h/dx), round(w/dx),)).T
    p, _ = s.to_planar(to_2D=to_2D)
    p = p.polygons_full

    # fig = plt.figure(1, )
    # ax = fig.add_subplot(121)
    # plot_polygon(p[0])
    # plt.show()

    p = clip_by_rect(GeometryCollection(p),
                     xmin, ymin, xmax, ymax)
    #  x-w/2, y-h/2, x + w/2, y + h/2)
    if not p:
        return None
    # plot_polygon(p,)
    # plt.show()

    p = shapely.affinity.translate(p, (xmax-xmin)/2-x, (ymax-ymin)/2-y)
    svg = s2svg(p, [0, 0, w, h], dx)
    svg2png(bytestring=svg, write_to='temp.png', background_color="#00000000",
            output_width=round(w/dx), output_height=round(h/dx))
    # p.plot()
    # with open('_.svg', 'w') as f:
    #     f.write(svg)
    # plot_polygon(p,)
    # plt.show()

    img = iio.imread('temp.png')
    img = np.sum(img, 2).T
    d = np.max(img)-np.min(img)
    b = np.min(img)
    if d == 0:
        if b == 0:
            d = 1
        else:
            d = b
            b = 0
    img = (img-b)/(d)
    # img = rasterio.features.rasterize(
    #     [p], transform=(dx, 0, 0, 0, dx, 0, 0, 0, dx), out_shape=(round(h/dx), round(w/dx),)).T
    # plt.imshow(img)
    # plt.show()

    return img


p = shapely.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
svg = s2svg(p, [0, 0, 2, 2], .2)
svg2png(bytestring=svg, write_to='temp.png', background_color="#00000000",
        output_width=10, output_height=10)


def material_slice(c, dx, center, w, h, normal, layers, layer_stack, materials=MATERIALS):
    LAYER_VIEWS["WGCLAD"].visible = True
    view = LAYER_VIEWS.layer_views["WGCLAD"]

    _layer_views = copy.deepcopy(LAYER_VIEWS)
    _layer_views.layer_views.clear()
    for i, layer in enumerate(layers):
        k = f"l{i}"
        _layer_views.layer_views[k] = copy.deepcopy(view)
        _layer_views.layer_views[k].layer = layer

    layers = sum([[[v.mesh_order, v.material, tuple(layer), k]
                 for k, v in get_layers(layer_stack, layer, withkey=True)] for layer in layers], [])

    epsmin = 100
    eps_array = 0
    layers = sorted(layers, key=lambda x: -x[0])
    for layer in layers:
        m = layer[1]
        k = layer[3]
        eps = materials[m].epsilon

        _layer_stack = copy.deepcopy(layer_stack)
        _layer_stack.layers.clear()
        _layer_stack.layers[k] = layer_stack.layers[k]
        scene = c.to_3d(
            layer_stack=_layer_stack,
            layer_views=_layer_views)
        # mesh = sum(scene.geometry.values())
        _mask = raster_slice(scene,
                             dx, w=w, h=h,
                             center=center,
                             normal=normal)
        # print(layer, center)
        if _mask is None:
            # print("no mask ")
            pass
        else:
            if eps < epsmin:
                epsmin = eps
            eps_array = _mask*eps+(1-_mask)*eps_array
    holes = np.where(eps_array < epsmin, 1, 0)
    eps_array = (1-holes)*eps_array+epsmin*holes
    # plt.clf()
    # plt.imshow(eps_array)
    # plt.show()
    return eps_array


def material_voxelate(c, dx, center, l, w, h,  normal, layers, layer_stack, materials):
    # print(l)
    return np.stack([material_slice(c, dx, [center[0]+x]+center[1:], w, h, normal,    layers, layer_stack, materials) for x in np.arange(dx/2, l-.001, dx)],)


def get_layers(layer_stack, layer, withkey=False):
    r = []
    for k, x in layer_stack.layers.items():
        l = x.layer
        if hasattr(l, "layer"):
            if tuple(l.layer) == tuple(layer):
                if withkey:
                    x = k, x
                # print(layer, k, x)
                r.append(x)
    if r:
        return r

    for k, x in layer_stack.layers.items():
        l = x.derived_layer
        if hasattr(l, "layer"):
            if tuple(l.layer) == tuple(layer):
                if withkey:
                    x = k, x
                # print(layer, k, x)
                r.append(x)
    return r

    # if hasattr(x.layer, "layer1") and x.layer.layer1.layer == layer:
    #     return x


def show_solution(path=None):

    if path is None:
        path = sorted(os.listdir(RUNS_PATH))[-1]
        path = os.path.join(RUNS_PATH, path)
    print(f"showing solution from {path}")
    # Load an image
    for s in ["after.png", "run1.png"]:
        try:
            img = Image.open(os.path.join(path, s))
        except:
            pass
        else:
            img.show()
            try:
                display(img)
            except:
                pass
            return
    print("no plot image found")



from .setup import *
from .constants import *
from .layers import *
from .utils import *
import gdsfactory as gf
from copy import deepcopy
import matplotlib.pyplot as plt
import numpy as np

from gdsfactory.cross_section import Section
from gdsfactory.generic_tech import LAYER_STACK, LAYER


def sparams_problem(c: gf.Component,
                    margin=None,  # zmargin=None,zlims=None,
                    dx=.05, wavelengths=[1.55], center_wavelength=None, keys=None,
                    approx_2D=False, layer_stack=LAYER_STACK,
                    study="sparams",
                    **kwargs):
    wavelengths = sorted(wavelengths)
    d = 2 if approx_2D else 3
    prob = dict()

    ports = [int(p.name[1])
             for p in c.get_ports_list(prefix="o")]
    if keys is None:
        io = {i: ports for i in ports}
    else:
        io = SortedDict()
        for k in keys:
            (o, i) = longname(k).split(",")
            if i not in io:
                io[i] = SortedSet()
            io[i].add(o)
        for i in io:
            if i not in io[i]:
                io[i].add(i)

    runs = [{
        "d": d,
        "sources": {
            port_number(i): {
                "center": (np.array(c.ports[f"o{port_number(i)}"].center)/1e3).tolist(),
                "width": (np.array(c.ports[f"o{port_number(i)}"].width)/1e3).tolist(),
                "normal": normal_from_orientation(c.ports[f"o{port_number(i)}"].orientation),
                # "endpoints": extend(c.ports[f"o{port_number(i)}"]., margin),
                "wavelength_mode_numbers": {w: [mode_number(i)] for w in wavelengths},
                "port": port_number(i),
            }},
        "monitors": {
            port_number(o): {
                "port": port_number(o),
                "normal": normal_from_orientation(c.ports[f"o{port_number(o)}"].orientation),
                "center": (np.array(c.ports[f"o{port_number(o)}"].center)/1e3).tolist(),
                "width": (np.array(c.ports[f"o{port_number(o)}"].width)/1e3).tolist(),
                # "endpoints": extend(c.ports[f"o{port_number(o)}"].endpoints, margin),
                "wavelength_mode_numbers": {w: [mode_number(o)] for w in wavelengths},
            } for o in io[i]},
    } for i in io]

    prob = setup(c, study=study,  dx=dx,
                 runs=runs, margin=margin,
                 layer_stack=layer_stack, approx_2D=approx_2D, **kwargs)
    prob["wavelengths"] = wavelengths

    return prob
